@@ -1,10 +1,10 @@
-#define RUBY_VERSION "2.2.4"
-#define RUBY_RELEASE_DATE "2015-12-16"
-#define RUBY_PATCHLEVEL 230
+#define RUBY_VERSION "2.3.6"
+#define RUBY_RELEASE_DATE "2017-12-14"
+#define RUBY_PATCHLEVEL 384
 
-#define RUBY_RELEASE_YEAR 2015
+#define RUBY_RELEASE_YEAR 2017
 #define RUBY_RELEASE_MONTH 12
-#define RUBY_RELEASE_DAY 16
+#define RUBY_RELEASE_DAY 14
 
 #include "ruby/version.h"
 
@@ -30,7 +30,10 @@
 #endif
 
 #if RUBY_REVISION
-# ifdef RUBY_BRANCH_NAME
+# if RUBY_PATCHLEVEL == -1
+#  ifndef RUBY_BRANCH_NAME
+#   define RUBY_BRANCH_NAME "trunk"
+#  endif
 #  define RUBY_REVISION_STR " "RUBY_BRANCH_NAME" "STRINGIZE(RUBY_REVISION)
 # else
 #  define RUBY_REVISION_STR " revision "STRINGIZE(RUBY_REVISION)
